/* -*- mode: c++ -*- */
#ifndef __DAQpp_Fifo_h__
#define __DAQpp_Fifo_h__
#include <DAQ++/Thread.h>
#include <DAQ++/exception.h>

namespace DAQpp
{
    class Fifo;
    struct FifoPriv;
    /**
     *  A Condition that notifies when data is available
     */
    class DataCondition : public Condition
    {
        private:
            /// Pointer to the fifo
            Fifo *fifo;
        public:
            /** Constructor.
             \param f a pointer to the constructor to spy.
             */
            DataCondition(Fifo *f) :
                fifo(f)
            {
            }

            /// The condition: the fifo is not empty
            bool condition();
    };

    /**
     * Thread to get data from a stream. This class will create a Thread to
     * poll for data from a stream.
     */
    class FifoThread : public Thread
    {
        private:
            /// Pointer to the Fifo
            Fifo *fifo;

        public:
            /// Constructor
            FifoThread(Fifo *f) :
                fifo(f)
            {
            }
            void run();
    };

    /**
     * \brief class FifO.
     * \ingroup IO
     *
     * This class implements a fifo. It can be used as a data
     * communication channel between different threads, for instance
     * the RunManager::writeData() and a data logger.
     *
     * It can be used in two modes:
     * - A pure FIFO, with 2 processes attached,
     *   one pushing in data and a second one polling out data
     * - One can attach a stream to the Fifo, this will create a thread that
     *   pushes stream data into the memory, and the user polls out the data
     *   from the other end. In order to attach an open stream one should call
     *   attach(int fd)
     */
    class Fifo : public Mutex
    {
        private:
            /// The data condition
            DataCondition condition;
            /// The thread to poll for data from an attached stream
            FifoThread thread;

            /// File descriptor of the stream
            int fd;

            /// Private structure
            FifoPriv *_priv;

            /// A state boolean telling if the thread is running
            //volatile bool _in_loop;

            /// A state boolean that will make the thread exit
            //volatile bool _end_poll;

            /// buffer size
            unsigned int bfsiz;

            /// buffer
            char *buff;

            /// read pointer
            char *rdptr;

            /// write pointer
            char *wrptr;

            /// End of the buffer
            char *end;

            /**
             * This function will increase the size of the buffer.
             * The final size of the Fifo will always be a power of 2.
             */
            void grow(int sz);

            /// Reads data, when available, from the attached stream.
            void read_from_channel();

            /// starts the thread when attaching a stream
            static void *start_loop(Fifo *);

            friend class FifoThread;
        public:
            /// constructor
            Fifo(unsigned int size);

            /// destructor
            ~Fifo();

            /// returns the size of the fifo
            unsigned int size() const
            {
                return bfsiz;
            }

            /// resizes the Fifo
            void resize(unsigned int sz);

            /// tells if the fifo is empty
            bool empty()
            {
                return rdptr == wrptr;
            }

            /// reads from the fifo
            int read(char *buf, int);
            
            /** \brief writes to the Fifo.
             *  \param buf A pointer to the data to write
             *  \param sz  amount of data to write
             *  \param force If force is false, the data will be lost when there is not enough
             *               space in the buffer. If force is true, the Fifo buffer will be
             *               enlarged to ensure that this block of data is not lost
             */
            int write(char *buf, int sz, bool force = false);
            /// resets the fifo
            void reset()
            {
                rdptr = wrptr = buff;
            }
            /// Returns the space left in the FIFO buffer
            int space_left();

            /// Returns the amount of data not yet read from the FIFO
            int data_available();

            /**
             * Waits until there is data available.
             * \param tmout An optional time out in microseconds.
             * \return true if there is data and false if there is a time out
             */
            int wait_for_data(int tmout=-1)
            {
                return condition.wait(tmout);
            }
            
            /// Attach the fifo to an open file
            int attach(int fd);
            /// Detach the fifo from the open file
            void detach();

            /// Tells if the Fifo thread is running
            bool running() const;

            /// Tells if the Fifo should stop polling for data from the open file
            bool endpoll() const;

            /// Make the Fifo stop polling for data from the open file
            void endpoll(bool x);

    };
}
#endif
