/* -*- mode: c++ -*- */
#ifndef __DAQpp_GTimer_included__
#define __DAQpp_GTimer_included__

#include <unistd.h>
#include <sys/time.h>
#include <cmath>

namespace DAQpp
{
    /**
     *  \brief class GTimer
     *  \ingroup Tools
     *
     *  This class implemets a timer. It is mainly used internally by the
     *  RunState objects.
     *
     *  \author Carlos Lacasta
     */
    class GTimer
    {
        private:
            /// start time of the timer
            struct timeval _start;
            /// stop time of the timer
            struct timeval _end;
            /// flags saying if the timer is active or not
            bool active;

        public:
            /// Constructor
            GTimer();
            /// Destructor
            ~GTimer()
            {}
            /// Starts the timer
            void start();
            /// Stops the timer
            void stop();
            /// resets the timer without stopping it
            void reset();
            /// returns the time in seconds elapsed since the start
            double operator()() const;
            /// returns true if the timer is active
            bool is_active() const
            {
                return active;
            }
            void set(double sec);
    };
}

inline DAQpp::GTimer::GTimer()
{
    start();
}
inline void DAQpp::GTimer::start()
{
    active = true;
    gettimeofday (&_start, NULL);
}
inline void DAQpp::GTimer::stop()
{
    active = false;
    gettimeofday (&_end, NULL);
}
inline void DAQpp::GTimer::reset()
{
    gettimeofday (&_start, NULL);
    _end = _start;
}
inline double DAQpp::GTimer::operator()() const
{
    double total;
    struct timeval elapsed;
    struct timeval end;

    if ( active )
        gettimeofday (&end, NULL);
    else
        end = _end;

    if (_start.tv_usec > end.tv_usec)
    {
        end.tv_usec += 1000000;
        end.tv_sec--;
    }
    elapsed.tv_usec = end.tv_usec - _start.tv_usec;
    elapsed.tv_sec  = end.tv_sec  - _start.tv_sec;

    total = elapsed.tv_sec + ((double) elapsed.tv_usec) / 1.e6;
    return total;
}

inline void DAQpp::GTimer::set( double s)
{
    struct timeval elapsed;
    elapsed.tv_sec = floor(s);
    elapsed.tv_usec = (s-elapsed.tv_sec)*1e6;
    gettimeofday (&_end, NULL);
    _start.tv_sec = _end.tv_sec - elapsed.tv_sec;
    _start.tv_usec = _end.tv_usec - elapsed.tv_usec;
     if ( _start.tv_usec < 0)
     {
         --_start.tv_sec;
         _start.tv_usec += 1000000;
     }
}

#endif
