/* -*- mode: c++ -*- */
#ifndef __DAQpp_Monitor_h__
#define __DAQpp_Monitor_h__
#include <deque>
#include <memory>
#include <DAQ++/Thread.h>

namespace DAQpp
{
    /**
     *  \brief   class Event.
     *  \ingroup Monitoring
     *
     *  The event class is just a buffer where the data of a given event
     *  can be stored. It is in fact the class that Monitor will use to
     *  build the buffer of event that can be used to monitor the run.
     *
     *  \author Carlos Lacasta
     */
    class Event
    {
        private:
            /// size of the buffer
            int _size;
            /// pointer to the buffer
            char *_data;

            /// used internally to implement copy, construction and assignment
            void cpy(const Event &e);

            /// The event allocator
            static std::allocator<Event> event_allocator;
            /// The data allocator
            static std::allocator<char> data_allocator;

            void free_data()
            {
                data_allocator.deallocate(_data, _size);
                _data = 0;
            }
        public:
            /// Constructor
            Event(int s, const char *d);
            /// Constructor
            Event(const Event &e) : _size(0), _data(0)
            {
                cpy(e);
            }
            /// Asignment operator
            Event &operator=(const Event &e)
            {
                if ( &e != this ) cpy(e);
                return *this;
            }

            /// Destructor
            ~Event();

            /// Returns the size of the data
            int size() const
            {
                return _size;
            }
            /// A pointr to the data buffer
            const char *data() const
            {
                return _data;
            }

            /// new is overloaded
            inline void* operator new(size_t);   /// overloaded new operator
            /// delete is overloaded
            inline void  operator delete(void*); /// overloaded delete operator

    };


    // overloaded new operator
    inline void* Event::operator new(size_t)
    {
        void* anEvent;
        anEvent = event_allocator.allocate(1);
        return anEvent;
    }

    // implementation of overloaded delete operator
    inline void Event::operator delete(void* anEvent)
    {
        event_allocator.deallocate(static_cast<Event *>(anEvent), 1);
    }

    struct MutexPriv_t;
    /**
     *   \ingroup Monitoring
     *
     *   class Monitor.
     *
     *   The Monitor class implements a buffer of Event objects. It is
     *   implemented as a queue in which new events will only be pushed if
     *   the current size (number of elements in the queue) is smaller than
     *   a specified maximum. The size limit can be set either
     *   on the constructor or changed afterwards.
     *
     *   It also derives from Thread and the monitoring of the sampled
     *   events should be done in the implementation of the run()
     *   method. In that method, the Monitor instance should use
     *   get_event() to get the next event in the Event queue. If the
     *   queue is empty, get_event() blocks.
     *
     *   get_event() accepts a timeout argument and it will wake up and
     *   return MonitorTimeOut instead of a valid pointer. Similarly,
     *   it will return MonitorDie when the monitoring should die.
     *
     *   \author Carlos Lacasta
     */
    class Monitor : public std::deque<Event *>, public Thread, public Condition
    {
        private:
            /// maximum number of events allowed in the buffer
            unsigned long mxsize;

            /// User loop get_event timeout
            int user_timeout;

            /// A private structure
            MutexPriv_t *_privm;

        public:
            /** This is a magic number that get_event() may return. If this
             *  is the case the Monitor should die.
             */
            static Event *MonitorDie;
            /** This is a magic number that get_event() may return. If this
             *  is the case the Monitor timed out waiting for new events
             *  in the queue
             */
            static Event *MonitorTimeOut;
            /**
             * Constructor.
             * \param sz maximum size of the buffer
             */
            Monitor(int sz = 50);
            /// Destructor
            ~Monitor();

            /**
             *  Adds a new Event in the buffer.
             *
             *  There is no local copy made on this call. The Monitor stores
             *  just the event pointer. The user will have to handle properly
             *  the pointer when he/she retrieves it with get_event.
             *
             */
            void push_event( Event *ev );

            /**
             *  Retrieves the oldest Event from the queue.
             *
             *  This call will put the calling thread to sleep when there is no
             *  data available. As soon as a new Event is added to the buffer
             *  the calling thread will wake up and continue.
             *
             *  \param timeout if >0, the routine does not block and will
             *                 wake up after the given number of
             *                 microseconds.
             *  \return A pointer to an event. There are two exceptions to
             *                 that. 1) MonitorDie: the monitor thread
             *                 should quit. 2) MonitorTimeOut: we gave a
             *                 time out parameter and that time elapsed
             *                 without receiveing any Event.
             *  \attention The event pointer returned should be deleted by the
             *  calling thread.
             *
             */
            Event *get_event(int timeout = -1);

            /// The wait condition
            bool condition();

            /// Sets the maximum size of the buffer
            void set_max_size(unsigned long s)
            {
                mxsize = s;
            }

            /// Returns the current maximum size of the object
            unsigned long get_max_size() const
            {
                return mxsize;
            }

            /// Returns true if the buffer is full
            bool full();

            /// Empties the buffer
            void clean();

            /**
             *  Starts a new thread.
             */
            void start_monitor();

            /**
             * User routine called when monitoring starts.
             */
            virtual void on_start() {}

            /**
             *  Stops the monitor thread started with start_monitor()
             */
            void stop_monitor();

            /**
             * User routine called when the monitor stops
             */
            virtual void on_stop() {}

            /**
             * This is the run method of the thread.
             * Monitor implements a default method that calls user_run
             * whenever there is data available in the buffer.
             */
            void run();

            /**
             * Returns the user time out.
             */
            int get_user_timeout() const { return user_timeout; }

            /**
             * Sets the user time out.
             *
             * When running with the default thread loop, the call to get_event has
             * a time out set by the value specified here (note that these are units
             * or microseconds)
             */
            void set_user_timeout(int val) { user_timeout = val; }

            /**
             * If the method returns false, the thread will stop
             */
            virtual bool user_run(Event *)
            {
                return true;
            }

    };

}

#endif
