/* -*- mode: c++ -*- */
#ifndef __DataFormat_h__
#define __DataFormat_h__

/*
 *  There are several blocks in the data stream. At the beginning of each
 *  file you should always find a StartOfRun record and a Pedestal
 *  record. Optionally, you may find a MapRecord, that is, the
 *  association between chips channels and pixels in a 2D sensor or
 *  ScanRecords, which define the scan
 *
 *  The format of the StartOfRun record is
 *
 *  Word
 *   0     0xCAFE1bbb   bbb= size of data block
 *   1     0xtttttttt   tttttttt - time as returned by time(0)
 *   2     0xrrrrrrrr   rrrrrrrr - run number
 *   3     0xnnnnnnnn   nnnnnnnn - number of modules
 *   4     0x0nniiiii   nn = number of chips iiiii= module id
 *   .
 *  bbb    0x00000bbb   size of data block
 *
 *
 *  The format of the Pedestal data is the following
 *     Word
 *     0     0xCAFEibbb   i= event type bbbb= size of data block
 *     1     0xaaaaaaaa   aaaaaaaa = module id
 *     2     0xaaaabbbb   aaaa= n chips bbbb= nchan
 *     3     0xnnnnpppp   nnnn = 50*noise, pppp=50*pedestal
 *     .     .
 *     bbbb  bbbb         size of data block
 *
 *
 *  The format of the MapRecord is:
 *     Word
 *     0     0xCAFEibbb   i= event type bbbb= size of data block
 *     1     0xaaaaaaaa   aaaaaaaa = module id
 *     2     0xiiiijjjj   iiii = nx, jjjj=ny
 *     2     0xaaaabbbb   aaaa= pix. no. of chan 0, bbbb=pix no. of chan 1
 *     .     .
 *     bbbb  bbbb         size of data block
 *
 *     where pixel numbers go like:
 *        pix. no. = y * nx + x
 *
 *
 *  The format of the ScanRecord is:
 *     Word
 *     0     0xCAFEibbb   i= event type bbbb= size of data block
 *     1     0xaaaaaaaa   number of variables
 *     2     0xaaaaaaaa   number of events per scan point
 *     3     0xiaaabbbb   i=value format aaaa=var type bbbb=number of points
 *     4     0xbbbbbbbb   first value
 *     5     0xbbbbbbbb   last value
 *     .     .
 *     bbbb  bbbb         size of data block
 *
 *     Words 3-5 appear once per variable.
 *     The value format indicates how the values are stored. If it is
 *     one they are stored as float numbers, otherwise as integers.
 *
 *
 *  After that you may find Data records
 *
 *  The format of the Data record is the following
 *  Word
 *   0     0xCAFEibbb   i= event type bbb= size of data block
 *   1     0xbbaaaaaa   bb=number of adjacent aaaaaaaa event counter
 *   2     0xiiitannn   iii=module id t=data type a= readout mode nnn=data size
 *   3     data word 0
 *   4     data word 1
 *   .     .
 *  bbbb    bbbb         size of data block
 *
 *  -------------------------------------------------------------------------
 *  Scans
 *
 *  Every scan has a record at the beggining of a new point. The
 *  structure of the  record is
 *  Word
 *   0     0xCAFEibbb   i= event type bbbb= size of data block
 *   1     0xbbbbbbnn   bbbbbb = number of events nn=number of variables
 *   2     0xtvvvvvvv   t var type, v=value
 *   .
 *  bbbb    bbbb         size of data block
 *
 *
 *  -------------------------------------------------------------------------
 *  ADC data block
 *
 *     Data Format
 *     0     0xCAFEtbbb   t = event type, bbb= size of data block
 *     1     0xaaaaaaaa   aaaaaaaa event counter
 *     2     0xiiitnnnn   iii=module id t=data type  nnnn=data size
 *     3     0xnnnnnnnn   number of buffers
 *     4     acdcnnnn     nnnn number of channels in this buffer
 *     5     data word 1
 *           .
 *     n     last data word
 *     n+1   ADC event counter
 *     n+2   acdcnnnnnn      
 *     .     .
 *    bbbb    bbbb         size of data block
 *
 *    The data words have the format explained in the v785 manual
 *    data type is ADCdata
 *  -------------------------------------------------------------------------
 *  FBTDC data block
 *
 *     Data Format
 *     0     0xCAFEtbbb   t = event type (2), bbb = size of data block
 *     1     0xaaaaaaaa   aaaaaaaa event counter
 *     2     0xiiitnnnn   iii=module id t=data type  nnnn=data size
 *     3     data word 0
 *     4     data word 1
 *     .     .
 *    bbbb    bbbb         size of data block
 *
 *    The data words have the format explained in the 1877 manual
 *    data type is FBTDCdata
 *
 *  -------------------------------------------------------------------------
 *  SIS3300 data block
 *
 *     Data Format
 *     0      0xCAFEibbb      i=record type,  bbb=ignored
 *     1      0xaaaaaaaa      aaaaaaaa=event counter
 *     2      0xiiitnnnn      iii=module id,  t=module type, nnnn=unused
 *     3      0xbbbbbbbb      bbbbbbbb=size of data block
 *     4      0xcsssssss      c=# of inputs, sssssss=sample size
 *     5      0x000(w|r)rrrr  w=wrap_bit (@ bit 19), rrrrr=ring pointer[17 bit]
 *     6      data word 0
 *     .         .
 *     x      0xbbbbbbbb      size of data block
 *
 *     Two options for data word: raw, proccessed
 *
 *     RAW:
 *     data word = 0x0uuu0vvv uuu=even ch (0,2,4,..,ccc-2) raw value (12 bits)
 *     vvv=odd ch (1,3,5,... ccc-1) raw adc value (12 bits)
 *     There are ssss<<4 words per channel pair and (c+1)/2 channel pairs
 *     giving in total N=(ssss<<4)*(c+1)/2 data words
 *
 *     Size of data block is bbbbbbbb=N+6. Since N can be larger than
 *     0xfff (=4096) instead of size(=bbb) the first word carries record type
 *     and size is given in the next data word.
 *
 *     The record type is BigDataRecord, the data type is SIS3300data
 *  -------------------------------------------------------------------------
 *  SecondDetector data block
 *
 *     Data Format
 *     0      0xCAFEibbb      i=record type  bbb=ignored
 *     1      0xaaaaaaaa      aaaaaaaa=event counter
 *     2      0xiiitnnmm      iii=module id t=data type nn=#amp mm=#modules
 *     3      0xbbbbbbbb      size of data block
 *     4      data word 0     begin 1st amplifier information // not implemented
 *     .         .
 *     .      SIS3300 data blocks
 *     .         .
 *     x      0xbbbbbbbb      size of data block
 *
 *     The record type is BigDataRecord, the data type is SecDetData
 *    At the end of each file, you should always find an EndOfRunRecord
 *    with the following format
 *
 *  Word
 *   0     0xCAFE3004   004 = size of data block
 *   1     0xtttttttt   tttttttt - time as returned by time(0)
 *   2     0xrrrrrrrr   rrrrrrrr - run number
 *   3     0x00000004   size of data block
 *
 *
 */

const unsigned int DataHeader = 0xcafe0000;

enum EventType
{
    StartOfRunRecord = 0x1000,
    DataRecord = 0x2000,
    EndOfRunRecord = 0x3000,
    PedestalRecord = 0x4000,
    NewPointRecord = 0x5000,
    MapRecord = 0x6000,
    ScanRecord = 0x7000,
    ConfigRecord = 0x8000,
    BigDataRecord = 0xa000,
};
enum DataType
{
    NormalData = 0x10000,
    PulseData = 0x20000,
    ScanData = 0x30000,
    ADCdata = 0x40000,
    CSPRINTdata = 0x80000,
    FBTDCdata = 0x90000,
    SIS3300data = 0xa0000,
    SecDetData = 0xb0000,
    TDCCaenData = 0xc0000,
    FPGACaenData = 0xd0000,

};

#define IsDataBlock(x)            (((x)&0xffff0000)==DataHeader)
#define IsStartOfRunRecord(x)     (((x)&0xf000)==StartOfRunRecord)
#define IsDataRecord(x)           (((x)&0xf000)==DataRecord)
#define IsEndOfRunRecord(x)       (((x)&0xf000)==EndOfRunRecord)
#define IsPedestalRecord(x)       (((x)&0xf000)==PedestalRecord)
#define IsNewPointRecord(x)       (((x)&0xf000)==NewPointRecord)
#define IsMapRecord(x)            (((x)&0xf000)==MapRecord)
#define IsScanRecord(x)           (((x)&0xf000)==ScanRecord)
#define IsBigDataRecord(x)        (((x)&0xf000)==BigDataRecord)
#define IsConfigRecord(x)         (((x)&0xf000)==ConfigRecord)

#define IsNormalData(x)           (((x)&0xf0000)==NormalData)
#define IsPulseData(x)            (((x)&0xf0000)==PulseData)
#define IsScanData(x)             (((x)&0xf0000)==ScanData)
#define IsADCdata(x)              (((x)&0xf0000)==ADCdata)
#define IsCSPRINTdata(x)          (((x)&0xf0000)==CSPRINTdata)
#define IsFBTDCdata(x)            (((x)&0xf0000)==FBTDCdata)
#define IsSIS3300data(x)          (((x)&0xf0000)==SIS3300data)
#define IsSecDetData(x)           (((x)&0xf0000)==SecDetData)
#define IsFPGACaenData(x)           (((x)&0xf0000)==FPGACaenData)

#endif
