/* -*- mode: c++ -*- */
#ifndef __ConfigParser_h__
#define __ConfigParser_h__

#include <string>
#include <sstream>
#include <algorithm>
#include <vector>
#include <map>
#include <ctype.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <config/ConfigException.h>
#include <config/SectionParser.h>
#include <config/SchemaPath.h>
#include <DAQ++/RunManager.h>

/**
   \addtogroup CONFIG

   This describes how the configuration file is handled.

   The program is configured with an xml file that follows the rules
   defined in daqvme.xsd

   It essentially contains 5 sections:
      - scans: predefined scans
      - paths: allows to define some useful paths
      - IO: settings for the file I/O operations of the program.
      - modules: defines the modules that will participate in the
                 acquisition.
      - plugins: defines the plugins to handle the module
                 configuration and the trigger handler

   The plugins section is controled by the ConfigParser, scans, paths
   and IO are controled by the default SectionHandlers:
      - PathConfig for the paths section
      - ScanConfig for the scans section
      - IOConfig   for the IO section

   The modules section should be controled by the handlers defined in
   the plugin section.

   <h2>The plugin part</h2>

   This is used to define how the modules listed in the modules
   section are handled. Each plugin corresponds to a module and
   specifies the library where the code exists and the name of the
   function in that library that will create the module handler. If no
   library is given, we will assume that the code lives in the program
   code.

   It also provides the possibility of defining a hook to handle
   trigger conditions and run VMEDAQ in interrupt mode

   An example follows
   \verbatim
   <plugins>
      <plugin name="MyModule" module="module">
         <lib>libMyModule.so</lib>
  <hook>create_MyModule</hook>
      </plugin>
      <interrupt>
         <enable>enable_interrupts</enable>
         <disable>disable_interrupts</disable>
         <lib>libCSprint.so</lib>
      </interrupt>
   </plugins>
   \endverbatim

   The plugin element defines the module to be handled \em
   MyModule. It should be defined somewhere in the configuration
   file. The function create_MyModule in the library libMyModule.so
   will return the pointer of a MyModule object that will be used in
   the acquisition. If lib is not given, the program assumes that the
   creation function is defined in the main program.

   The interrupt element defines the hooks that will enable and
   disable external triggers or, in other words, trigger interrupts in
   the software.

   The element enable is the name of a routine that will take the
   necessary actions to enable detection of trigger conditions. This
   function should return 0 if things went OK and -1 otherwise.

   disable is the name of the routine that will disable that
   behaviour.

   Like in plugin, lib specifies the path of the library where those
   two routines are defined. If no library path is given, the program
   assumes the function is defined in the main

   <h2>The scans section</h2>
   This section is used to configure possible scans. Each scan
   has a list of variables for which a loop will be made.

   Also specify here is the number of events for each scan point.

   \verbatim
   <scans nevts="1000">
      <scan name="gain">
         <type name="pulse" from="0.1" to="4." step=".1"/>
         <type name="threshold" from="0" to="25" step="1"/>
      </scan>
   </scans>
   \endverbatim

   In the previous example we define a single scan named \em gain of
   two variables: pulse and threshold. The possible variables to scan
   are:
      - pulse
      - threshold
      - channel
      - trim

   The nevts attribute indicates the number of events per scan point.
   The meaning of from, to and step are quite self-explaining.

   <h2> The paths section</h2>
   This section allows to define different paths that may be handy
   for the program. In particular the one named glade is used to
   define the path where the glade files will be searched.

   \verbatim
   <paths>
      <path name="data" path="/data"/>
      <path name="glade" path="/home/lacasta/test/gui/glade"/>
   </paths>
   \endverbatim

   <h2> The IO section </h2>
   This section allows to define the main IO options of the DAQ
   program, like the output path, the compression level, etc
 */




class ModuleParser;


/**
 *   \brief class ConfigParser.
 *   \ingroup CONFIG
 *
 *   This is the class reponsible of parsing the configuration file. It
 *   handles all the internals of the xml parser. It works by mantaining
 *   a map of SectionParser objects that will parse the different
 *   sections. Some of the sections, however, are handled by
 *   ConfigParser itself. One of these sections is plugins.
 *
 */
class ConfigParser
{
    private:
        /// The document
        xmlDoc *doc;

        /// The vmedaq element
        xmlNodePtr vmedaq;
        
        /** The pathlist where to find the configuration files.
         *  The syntax is the usual one: paths separated by ':'
         */
        std::string config_path;

        /** The pathlist where to find the schemas
         */
        SchemaPath finder;

        /** Flag telling if we want to validate the configuration file.
         */
        bool do_validate;

        typedef std::map<std::string, SectionParser *> HandlerMap;
        typedef HandlerMap::iterator HandlerIterator;
        
        /// This is the list of SectionParser objects
        HandlerMap handlers;

        typedef std::map<int, ModuleParser *> IdHandler;
        typedef IdHandler::iterator IdHandlerIterator;
        /// This is the list of ModuleParser objects
        IdHandler id_handler;

        // Navigates the document tree
        void navigate(xmlNodePtr);

        // Process one node
        void process_node(xmlNodePtr);

    protected:
        // Hide constructors
        ConfigParser();
        ConfigParser(const ConfigParser &);
        void operator=(const ConfigParser &);

        void clean();

    public:
        // destructor
        ~ConfigParser();

        void reset();
        bool create_doc();
        /// returns a pointer to the unique instance
        static ConfigParser *theConfig()
        {
            static ConfigParser cfg;
            return &cfg;
        }

        /// Returns the configuration path
        std::string get_config_path() const
        {
            return config_path;
        }
        void set_config_path(const std::string &s)
        {
            config_path = s;
        }
        void add_config_path(const std::string &s)
        {
            if (config_path.empty())
                config_path = s;
            else
            {
                config_path += (":" + s);
            }
        }

        /// Adds a path in the schema path
        void add_schema_path(const std::string &s)
        {
            finder.add_path(s);
        }

        /// Request or prevent validation
        void set_do_validate(bool x)
        {
            do_validate = x;
        }

        /// parses the file
        bool parse(const char *file);

        /// validates the file
        bool validate(const char *dir = 0);

        /// saves the current configuration
        int save_config(const char *ofile);

        /// returns the root node of the document
        xmlNodePtr root()
        {
            return doc ? xmlDocGetRootElement(doc) : 0;
        }

        /**
         * Returns the paths node
         */
        xmlNodePtr get_paths_node();
        
        /**
         * Returns the IO node
         */
        xmlNodePtr get_io_node();
        
        /**
         * Returns the general node
         */
        xmlNodePtr get_general_node();
        
        /**
         * Returns the scans node
         */
        xmlNodePtr get_scans_node();
        
        /** 
         * Returns the plugins node
         */
        xmlNodePtr get_plugins_node();

        /**
         * Returns the modules node
         */
        xmlNodePtr get_module_node();
        
        /**
         * Find and element in a given node
         */
        xmlNodePtr find_element(xmlNodePtr node, const std::string &name);
        xmlNodePtr find_element(xmlDoc *, const std::string &name);
        /** Get a list of nodes matching the XPath expresion at the input
         *  The path should be relative to the root of the document
         */
        int get_node(std::string, std::vector<xmlNodePtr> &);

        /** Get a list of nodes matching the XPath expresion at the input
         *  The path should be relative to the given node
         */
        int get_node(std::string, xmlNodePtr node, std::vector<xmlNodePtr> &);

        /// Gets the content of a node
        std::string get_node_value(xmlNodePtr node);

        /// Gets the value of a node
        template <class T> T get_value(xmlNodePtr node)
        {
            T v;
            std::istringstream ss(get_node_value(node) );
            ss >> v;
            return v;
        }

        /** Gets the value of a node's child named name
         Assumes that there is only one...
         */
        template <class T> T get_value(xmlNodePtr node, const std::string &name)
                throw (ConfigException )
        {
            int nval;
            std::vector<xmlNodePtr> vval;
            nval = get_node(name, node, vval);
            T v;

            if (nval)
            {
                v = get_value< T >(vval[0]);
            }
            else
            {
                std::ostringstream ostr;
                ostr << "Element " << node->name << "/" << name << " does not exist";
                throw ConfigException( ostr.str() );
            }
            return v;
        }

        /**
         * Removes the given node
         */
        void remove_child(xmlNodePtr node);
        
        /**
         * Remove the given child of the node. 
         * It will only remove the first occurence.
         * \returns true if something was removes and false otherwise
         */
        bool remove_child(xmlNodePtr node, const std::string &child);
        
        /// Creates a new element
        template <class T> xmlNodePtr create_child(xmlNodePtr node, const std::string &name,
                                             const T&val)
        {
            std::ostringstream os;
            os << val;
            return xmlNewTextChild(node, NULL, (xmlChar *)name.c_str(), (xmlChar *)os.str().c_str() );
        }

        /// Sets the value of a node
        template <class T> void set_value(xmlNodePtr node, const T &val)
        {
            std::ostringstream os;
            os << val;
            xmlNodeSetContent(node, (xmlChar *)os.str().c_str());
        }

        /// Sets the value of a node's child named name
        template <class T> void set_value(xmlNodePtr node, const std::string &name,
                                          const T &val)
        {
            std::ostringstream os;
            os << val;

            int nval;
            std::vector<xmlNodePtr> vnode;
            nval = get_node(name, node, vnode);
            if (nval)
            {
                set_value(vnode[0], val);
            }
            else
            {
                create_child(node, name, val);
            }
        }

        /// If the value of the node is different than val,
        /// replace by val
        template<class T> void check_value(xmlNodePtr node, const T &val)
        {
            try
            {
                T old = get_value<T>(node);
                if ( val != old )
                set_value<T>(node, val);
            }
            catch ( ConfigException &e )
            { 
                // the element does not exist
                set_value(node, val);
            }
        }

        /// If the value of the node's child named name
        //  is different than val, replace it by val
        template<class T> void check_value(xmlNodePtr node, const std::string &name,
                                           const T &val)
        {
            try
            {
                T old = get_value<T>(node, name);
                if ( val != old )
                {
                    set_value<T>(node, name, val);
                }
            }
            catch ( ConfigException &e )
            { 
                // the element does not exist
                set_value<T>(node, name, val);
            }
        }

        /// Get the value of an attribute
        template <class T> T get_attribute(xmlNodePtr node, const std::string &name)
                throw(ConfigException )
        {
            xmlChar *val = xmlGetProp(node, (xmlChar *)name.c_str());
            if (val == 0)
                throw ConfigException("attribute " + name + " not present");

            std::istringstream os( (const char *)val);

            T out;
            os >> out;
            xmlFree(val);
            return out;
        }

        /// Sets the value of an attribute
        template<class T> void set_attribute(xmlNodePtr node, const std::string &att,
                                             const T &val)
        {
            std::ostringstream ostr;
            ostr << val;
            xmlSetProp(node, (xmlChar *)att.c_str(), (xmlChar *)ostr.str().c_str());
        }

        template<class T> void check_attribute(xmlNodePtr node, const std::string &att,
                                               const T &val)
        {
            try
            {
                T old = get_value<T>(node, att);
                if ( val != old )
                set_attribute(node, att, val);
            }
            catch ( ConfigException &e )
            {
                set_attribute(node, att, val);
            }
        }

        /// Gets and interprets the content of a node's attribute as a bit array
        int get_bit_array(xmlNodePtr node, const std::string &nam, int nx,
                          std::vector<int> &vv);

        /// Gets and interprets the content of a node as a bit array
        int get_bit_array(xmlNodePtr node, int nx, std::vector<int> &vv);

        /// Gets the content of a node's attribute as an array
        template <class T> void get_array(xmlNodePtr node, const std::string &nam,
                                          std::vector<T> &v)
        {
            v.clear();
            T i;
            try
            {
                std::istringstream ss( get_attribute<std::string>(node, nam) );
                while (1)
                {
                    ss >> i;
                    if ( !ss )
                    break;
                    v.push_back(i);
                }
            }
            catch ( ConfigException &e )
            {
                throw;
            }
        }

        /// Gets the content of a node as an array
        template <class T> void get_array(xmlNodePtr node, std::vector<T> &v)
        {
            T i;
            v.clear();
            try
            {
                std::istringstream ss( get_node_value(node) );
                while (1)
                {
                    ss >> i;
                    if ( !ss )
                        break;
                    v.push_back(i);
                }
            }
            catch (ConfigException &e)
            {
                throw;
            }
        }

        /// Registers a handle
        void register_handler(const std::string &s, SectionParser *t);

        /// Register an id handler
        void register_id_handler(int, ModuleParser *t);

        /// returns the handler of section name
        SectionParser *get_handler(const std::string &name)
        {
            HandlerIterator ip;
            ip = handlers.find(name);
            if (ip == handlers.end())
                return 0;
            else
                return ip->second;
        }

        /// returns the handler of section name
        ModuleParser *get_id_handler(int name)
        {
            IdHandlerIterator ip;
            ip = id_handler.find(name);
            if (ip == id_handler.end())
                return 0;
            else
                return ip->second;
        }

        xmlNodePtr find_plugin(const std::string &name);
        int        remove_plugin(const std::string &name);
        xmlNodePtr find_module(const std::string &element, const std::string &id);
        int        remove_module(const std::string &element, const std::string &id);
};

template <>
bool ConfigParser::get_value(xmlNodePtr node);

template <>
void ConfigParser::set_value(xmlNodePtr node, const bool &val);

template <>
xmlNodePtr ConfigParser::create_child(xmlNodePtr node, const std::string &name, const bool &val);

template <>
void ConfigParser::set_attribute(xmlNodePtr node, const std::string &att, const bool &val);

template <>
bool ConfigParser::get_attribute(xmlNodePtr node, const std::string &name)
throw( ConfigException );

template <>
std::string ConfigParser::get_attribute(xmlNodePtr node, const std::string &name)
throw( ConfigException );

#endif
