#ifndef __PluginParser_h__
#define __PluginParser_h__

#include <map>
#include <config/SectionParser.h>
#include <config/SearchPath.h>
#include <DAQ++/RunManager.h>
#include <DAQ++/FileIOManager.h>
#include <config/PluginException.h>

class Hook;
typedef DAQpp::RunManager *(*RunManagerCreator)(const std::string &, const std::string &);
typedef DAQpp::FileIOManager *(*LoggerCreator)(bool, int);

/**
 * \brief class PluginParser
 * \ingroup CONFIG
 * 
 * This is a specialized desdendant of SectionParser used to handle
 * plugins in the plugins section. 
 * 
 */
class PluginParser : public SectionParser
{
    private:
        typedef std::map<std::string, RunManagerCreator> RunManagerFactory;
        RunManagerFactory run_manager_factory;

        int (*trigger_enable)();
        void (*trigger_disable)();
        void (*trigger_reset)();
        void (*trigger_start)();
        void (*trigger_stop)();
        LoggerCreator data_logger;
        
        /// List of opened libraries
        class LibMap : public std::map<std::string, void *>
        {
            public:
                LibMap() {}
                virtual ~LibMap()
                {
                    reset_libs();
                }
                
                void reset_libs();
        };
        static LibMap libmap;

    public:
        PluginParser();
        virtual ~PluginParser();

        void parse_node(xmlNodePtr node) throw (ConfigException );

        void parse_datalogger(xmlNodePtr node) throw (ConfigException );
        void parse_plugin(xmlNodePtr node) throw (ConfigException );
        void parse_interrupt(xmlNodePtr node) throw (ConfigException );
        void parse_run_manager(xmlNodePtr node) throw (ConfigException );

        /// Loads a library
        static void *load_library(const std::string &lib)
            throw (PluginException);
        
        /// Loads a symbol from a library
        static Hook load_symbol(const std::string &symbol, const std::string &lib="")
            throw (PluginException);
        
        
        /// Calls the fuction to enable trigger interrupts
        int enable_interrupts()
        {
            return (*trigger_enable)();
        }

        /// Calls the function to disable trigger interrupts
        void disable_interrupts()
        {
            (*trigger_disable)();
        }

        /// Calls the function to reset the trigger interrupts
        void reset_interrupts()
        {
            (*trigger_reset)();
        }

        /// Become sensitive to trigger interrupts
        void start_interrupts()
        {
            (*trigger_start)();
        }

        /// Ignore trigger interrupts, although they are still enabled
        void stop_interrupts()
        {
            (*trigger_stop)();
        }
        
        void reset_plugins();
        
        void set_interrups(const std::string &lib, 
                           const std::string &enable,
                           const std::string &disable,
                           const std::string &start,
                           const std::string &stop,
                           const std::string reset);
        
        void set_data_logger(const std::string &lib, const std::string &hook);
        void set_runmanager(const std::string &type, const std::string &lib, const std::string &hook);
        void set_plugin(const std::string &name, const std::string &module, const std::string &lib, const std::string &hook);
        
        /// Creates a RunManager.
        DAQpp::RunManager *create_run_manager(const std::string &type,
                                              const std::string &name);
        
        /// Creates a DataLogger
        DAQpp::FileIOManager *create_logger(bool, int);

        static void reset_libs() 
        {
            libmap.reset_libs();
        }
};

/**
 * This class is a wrapper of a symbol handler as returned by
 * dlopen
 */
class Hook
{
    public:
        typedef ModuleParser *(*HookPter)();
        typedef int (*IntFunc)();
        typedef void (*VoidFunc)();
    private:
        /** awful hack to cheat the compiler and store the void pointer
         *  returned by dlsym
         */
        union {
            void *vptr;
            HookPter tptr;
            LoggerCreator lptr;
            IntFunc  iptr;
            VoidFunc vdptr;
            RunManagerCreator rptr;
        } patch;

        void cpy(const Hook &h)
        {
            patch.vptr = h.patch.vptr;
        }
    public:
        /// constructor
        Hook(void *ptr=0)
        {
            patch.vptr = ptr;
        }
        
        Hook(const Hook &h)
        {
            cpy(h);
        }
        /// destructor
        ~Hook();

        /// Checks that it is a valid hook
        bool operator!() const
        {
            return patch.vptr ? false : true;
        }

        /// Complementary of operator!()
        operator bool()     const
        {
            return patch.vptr ? true : false;
        }

        /// Calls the hook with no arguments
        ModuleParser *operator()()
        {
            return patch.vptr ? (*patch.tptr)() : 0;
        }

        operator RunManagerCreator() const
        {
            return patch.rptr;
        }

        /// returns the pointer to the hook function returning a ModuleHandler
        operator HookPter() const
        {
            return patch.tptr;
        }

        /// returns the pointer to the hook that returns an int
        operator IntFunc()  const
        {
            return patch.iptr;
        }

        /// returns the pointer to a void hook
        operator VoidFunc() const
        {
            return patch.vdptr;
        }
        
        /// returns the pointer to a LoggerCreator
        operator LoggerCreator() const
        {
            return patch.lptr;
        }
        /// closes and frees up all the libraries
        static void close();
};



#endif /*__PluginParser_h__*/
