/* -*- mode: c++ -*- */
#ifndef __SectionParser_h__
#define __SectionParser_h__
#include <map>
#include <string>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <config/ConfigException.h>

/**
 *   \brief class SectionParser.
 *   \ingroup CONFIG
 * 
 *   This class is used by the ConfigParser to handle different sections
 *   of the configuration file. It should provide to methods:
 *      - parse_node:  this method will parse the corresponding section
 *                     and store locally the configuration parameters
 *       defined in there.
 *      - save_config: this method will be called if the configuration
 *                     is to be saved to an external file. The idea is
 *       that the class will check which parameters have
 *       changed since the file was loaded and update them
 *       in the document before saving.
 * 
 *   The SectionParser objects have to be registered in the
 *   ConfigParser. ConfigParser mantains a map with all the registered
 *   SectionParser objects. The keys of the map are the names of the
 *   different sections in the configuration file.
 */
class SectionParser
{
    protected:
        xmlNodePtr node;
        std::string id;
        std::string name;

        // To store the children nodes
        typedef std::map<std::string, xmlNodePtr> NodeMap;
        typedef NodeMap::iterator NodeMapIterator;

        NodeMap nmap;
    public:
        SectionParser() : node(0), id("")
        {}
        /// destructor
        virtual ~SectionParser()
        {
            nmap.clear();
        }

        const std::string get_id() const
        {
            return id;
        }
        void set_id(const std::string &s)
        {
            id = s;
        }

        /**
         * This method is called from ConfigParser to parse the associated
         * section of the config file
         */
        virtual void parse_node(xmlNodePtr node)
        throw ( ConfigException )
        {
            if (node == 0)
                return;

            return ;
        }

        /**
         * An alternative to re-parse a node. It assumes that one
         * has a way of identifying the section with a string.
         */
        virtual void parse_node_id(const std::string &id)
        throw (ConfigException)
        {}

        /**
         * Called to update the configuration.
         * This method should update all the nodes corresponding to
         * this section
         */
        virtual void save_config()
        throw (ConfigException)
        {}

        virtual void save_config_node(xmlNodePtr)
        throw (ConfigException)
        {}

        /**
         * This one will only save the section that can be identified
         * with id
         */
        virtual void save_config_id(const std::string &id)
        throw (ConfigException)
        {}

        std::string node_name() const
        {
            if (node)
                return std::string((char *)node->name);
            else
                return std::string("<null>");
        }

        int node_line() const
        {
            return xmlGetLineNo (node);
        }

        void set_node( xmlNodePtr _n)
        {
            node = _n;
            name = node_name();
        }

        std::string get_name() const
        {
            return name;
        }
};


#endif
