#ifndef __PulseGenerator_h__
#define __PulseGenerator_h__

#include <string>

/**
 * \ingroup DAQ
 * 
 * This class provides an interface to a pulse generator. 
 * It assumes that the pulse genertor can produce either 
 * single pulses or bursts of equal pulses. The user should 
 * implement a concrete instance of this interface in order 
 * to control a "real" pulse generator
 * 
 */
class PulseGenerator
{
    public:
        enum TriggerType
        {   Immediate, External, Bus};
    protected:
        /// Tells if we are in burst or single pulse mode
        bool burst;

        // Error string
        std::string error_string;

        // checks the status of the device
        virtual bool check_status()
        {
            return true;
        }
    public:
        /// Constructor
        PulseGenerator() :
            burst(false)
        {
        }

        /// Destructor
        virtual ~PulseGenerator()
        {
        }

        /// Returns the error string
        std::string get_error() const
        {
            return error_string;
        }

        /// Activates Burst mode
        virtual void activate_burst(bool z)
        {
            burst = z;
        }

        /// Returns if it is in burst mode
        bool is_burst() const
        {
            return burst;
        }

        /// Resets the device
        virtual void Reset()
        {
        }

        // Check initialization
        virtual bool operator!() const
        {
            return true;
        }

        /** 
         *  Defines the pulse shape.
         *  \verbatim
         * 
         *      +----------+              +------ . . . .
         *      |          |              |
         *      |          |              |
         *   ---+          +--------------+
         *     <>         <>
         *     Rise time  Fall time
         *
         *      <-- width ->
         *      <------------Period  ----->
         *  \endverbatim
         */
        virtual void define_pulse(double period, double width, double raise)
        {
        }

        /// Sets Output amplitude
        virtual void set_pulse(double amplitude)
        {
        }

        /// Sets the trigger frequency
        virtual void set_frequency(double freq)
        {
        }

        /// Defines the burst
        virtual void define_burst(double delay, double freq, int npulse, double period)
        {
        }

        /** 
         * \brief Triggers the pulse.
         * It is expected that if burst is false it will produce a 
         * single pulse, and when true, it will produce a burst of 
         * equal pulses
         */
        virtual void trigger()
        {
        }

        /**
         * Sets the delay of the pulse w.r.t the external
         * trigger of the generator.
         * 
         */
        virtual void set_delay(double d)
        {
        }

        /**
         * Sets the trigger type
         */
        virtual void set_trigger_type(TriggerType v)
        {
        }

        /// Enables the output
        virtual void enable_output()
        {
        }

        /// disables the output
        virtual void disable_output()
        {
        }
};
#endif
