/* -*- mode: c++ -*- */
#ifndef __ScanRunManager_h__
#define __ScanRunManager_h__

#include <daq/PulseRunManager.h>
#include <daq/ScanPoint.h>
#include <daq/NewPoint.h>
#include <DAQ++/GTimer.h>

/**
 * \brief class ScanRunManager
 * \ingroup DAQ
 *
 * This class handles scans. It is generic enough so that it can be used
 * as a base class for any custom RunManager that has to handle with
 * varying parameters.
 *
 * It adds 3 virtual methods that can be overloaded and that serve as hooks
 * in some key moments of the DAQ process:
 *    - init_run: called after PrepareForRun
 *    - new_trigger: called right after Trigger
 *    - new_point: called every time we change to a new point
 *
 */
class ScanRunManager : public PulseRunManager
{
    private:
        /// number of events (or seconds) a single point should last
        int nevt;

        /// interpret nevt as number of events or time
        bool _time;

        /// scan point counter
        int ipoint;

        /// list of variables
        std::vector<ScanVar> vvar;

        /// list of points in the scan
        std::vector<ScanPoint> sclist;

        /// the current point
        std::vector<ScanPoint>::iterator current_point;

        void set_values();


        /** This member will have the responsibility to change
         *  from one point to the next
         */
        NewPoint *next_point;
        NewPoint *own_handler;
    public:
        // constructor
        ScanRunManager(DAQpp::DAQid id);

        // destructor
        virtual ~ScanRunManager()
        {
            delete next_point;
        }

        void set_scan_list(int n, const std::vector<ScanVar> &v);
        void PrepareForRun();
        int  writeData(int size, const char *data);
        void Trigger();
        void EndOfRun();
        void write_header();
        bool CheckEndOfRun();

        bool is_time() const
        {
            return _time;
        }
        int get_n_points() const
        {
            return sclist.size();
        }
        int get_n_var() const
        {
            return vvar.size();
        }

        void is_time(bool x)
        {
            _time = x;
        }

        const ScanPoint &get_point(int x) const
        {
            return sclist[x];
        }
        const ScanVar &get_var(int x) const
        {
            return vvar[x];
        }
        const ScanPoint &get_current() const
        {
            if (current_point == sclist.end())
                return *sclist.begin();
            else
                return *current_point;
        }
        int nevt_per_point() const
        {
            return nevt;
        }

        /**
         * Sets the current parameter values in the Modules
         */
         void set_new_values();

        /// Called right after PreareForRun
         virtual void init_run() {}

         /// Called after Trigger
         virtual void new_trigger() {}

         /// Called every time we change to a new point.
         virtual void new_point() {}

         /**
          * Writes the ScanRecord right after the file header
          */
         void  writeHeaderXtra(std::ostream &);

         /**
          * Changes the NewPoint handler. If set to NULL, it will set
          * the default behaviour
          */
         void set_next_point_handler(NewPoint *np);
};

inline bool ScanRunManager::CheckEndOfRun()
{
    // Last scan point and last event of it
    if ( current_point == sclist.end() )
        return true;
    else
        return false;
}


#endif
