/* -*- mode: c++ -*- */
#ifndef __VMEmanager_h__
#define __VMEmanager_h__
#include <exception>
#include <vector>
#include <DAQ++/Thread.h>

/**
 * The below is there just for debugging dummy handlers
 */
#include <daq/random.h>

class VMEcommand;
/**
 *    \brief class VMEmanager
 *    \ingroup DAQ
 *    This class defines the interface to communicate with VME.
 *    This is an abstract class. The only abstract method is
 *    connected. All the rest are just dummies that do nothing.
 *
 *    connected has to be implemented by the child and should return true
 *    if it is a real handler or false if it is not connected to any VME
 *    bus
 *
 *    A VME driver will be loaded by the configuration parser with the
 *    usual `pluging' element in the configuration file. So, one has to
 *    follow similar steps as for the modules.
 *       - Create a shared library
 *       - Define a function that returns a pointer to your class like in
 *
 *       \verbatim
 *       class MyVME : public VMEmanager {
 *          private:
 *             ...
 *          public:
 *             ...
 *             bool connected() const { return true; } // true if real
 *
 *        };
 *
 *        extern "C" {
 *           VMEmanager *create_my_manager() {
 *              static MyVME manager;
 *              return &manager;
 *           }
 *        }
 *       \endverbatim
 *
 *       The plugin in the configuration file should be put in the
 *       general section, like
 *
 *       \verbatim
 *       <general>
 *          <pedestals> ... <\pedestals>
 *          <daq>
 *             <!-- Set the depth of the monitor buffer -->
 *             <monitor_buffer>150<\monitor_buffer>
 *             <!-- Plugin for the VME driver -->
 *             <vme_driver>
 *                <lib>libMyVME.so</lib>
 *                <hook>create_my_vme</hook>
 *             <\vme_driver>
 *          <\daq>
 *       <\general>
 *
 *       \endverbatim
 */
class VMEmanager
{
    public:
        /**
         * This structure is used to pass information when acknowlegeing a VME
         * interrupt.
         */
        typedef struct
        {
            unsigned long irqs;
            unsigned long vector;
            int level;
        }
        IRQinfo;

        /// Signature of an interrupt function
        typedef void (*InterruptFunc)(int);

        /// Signature of a factory
        typedef VMEmanager *(*VME_Factory)();

    protected:
        /** A mutex to have control of the access to the bus
         *  Sometimes the drivers have something like that implemented,
         *  but one never knows.
         */
        DAQpp::Mutex mtx;

        /** This varible will store the return code of the last VME
         * operation.
         */
        int _rc;

        /** A pointer to the factory function.
         *  This pointer is set by set_factory
         */
        static VME_Factory __factory__;

        /**
         * No constructors allowed. One should get the pointer of an
         * implementation from somewhere else (a factory, etc.)
         */
        VMEmanager()
            : _rc(0)
        {
        }

        VMEmanager(const VMEmanager &);
        void operator=(const VMEmanager &)
        {
        }

    public:
        virtual ~VMEmanager()
        {
        }

        static VMEmanager* VME()
        {
            return __factory__();
        }

        /**
         * set_factory
         *
         * This function sets the factory function to be the routine
         * with name routine in the library lib.
         *
         *   @param lib : the path of the libray
         *   @param routine: the routine of the library
         *
         */
        static void set_factory(const std::string &routine, std::string lib = "");

        /**
         * set_default_factory
         *
         * This function sets a dummy factory
         *
         */
        static void set_default_factory();


        /**
         *  Transforms the "vme address" into the address in the memory
         *  mapped by the driver. If no direct access to VME is allowed
         *  in the driver, then it should return the VME address.
         *
         *   @param add : vme address to map
         *   @param size : size of memory chunck to map
         *   @param type : type of address (A24, A32, ...)
         */
        virtual unsigned long map_address( unsigned long add, int size, int type)
        {
            return add;
        }

        /**
         * This method tells whether this is a real handler connected to
         * the bus or not.
         */
        virtual bool connected() const = 0;


        /** This method sets the VME status in the default implementation.
         *  It could eventually signal a VME error.
         */
        virtual void check_rc(int x)
        {
            _rc = x;
        }

        /// Returns the current VME status
        virtual int get_rc() const
        {
            return _rc;
        }

        // methods to read and write (A24)
        virtual unsigned char read8(unsigned long add )
        {
            return 0;
        }
        virtual void          write8(unsigned long add, unsigned char val)
        {
        }

        virtual unsigned short read16(unsigned long add )
        {
            return 0;
        }
        virtual void           write16(unsigned long add, unsigned short val)
        {
        }

        virtual unsigned long read32(unsigned long add )
        {
            return 0;
        }
        virtual void          write32(unsigned long add, unsigned long val)
        {
        }

        virtual unsigned long readA32D32(unsigned long add )
        {
            return 0;
        }
        virtual void          writeA32D32(unsigned long add, unsigned long val)
        {
        }

        // returns number of (long)words read
        virtual unsigned long readA32BLT32(unsigned long add, u_int32_t *dat, unsigned long nwords)
        {
            return 0;
        }
        virtual unsigned long readA322EVME(unsigned long add, u_int32_t *dat, unsigned long nwords)
        {
            return 0;
        }

        // stack commands
  virtual int stack_execute(std::vector<VMEcommand> &stack, std::vector<unsigned int> &data, int ntry=1);

        virtual int readA24_fifo32(unsigned long add, const int nwords, unsigned int *dat)
        {
            for (int i=0;i<nwords;i++) dat[i]=(unsigned int)read32(add);
            return 0;
        }

        // methods to handle interrupts
        virtual void acknowledge(IRQinfo *)
        {
        }
        virtual void Reset()
        {
        }
        virtual void enable_interrupts(InterruptFunc, unsigned short)
        {
        }
        virtual void disable_interrupts()
        {
        }

        // Acquire and release the MuTeX
        void acquire()
        {
            mtx.acquire();
        }
        void release()
        {
            mtx.release();
        }

};

class VMEcommand
{
public:
  enum CommandType {
    read8,write8,
    read16,write16,
    read32,write32,
    readA32D32,writeA32D32,
  };

  VMEcommand(CommandType type, unsigned long add, unsigned int data=0)
    : _type(type),
      _add(add),
      _data(data)
  { }
  ~VMEcommand() { }

  CommandType type() { return _type; }
  unsigned long add() { return _add; }
  unsigned int data() { return _data; }

private:
  VMEcommand(){}
  CommandType _type;
  unsigned long _add;
  unsigned int _data;
};

#endif
