#!/usr/bin/env python
"""
    This is an event viewer
"""

import os
import sys
import vmedaq
from optparse import OptionParser
import EventBuilder

# A few constants
(StartOfRun, PedestalRecord, MapRecord, ConfigRecord,
 ScanRecord, NewPointRecord, DataRecord, BigDataRecord 
 ) = tuple([ 1 << x for x in range(0, 8)])


#
#Check that we have root
#
__root_sys = os.getenv("ROOTSYS", None)
if not __root_sys:
    print "ROOTSYS not defined. Quitting"
    sys.exit(1)
    
__root_path = os.path.join(__root_sys, "lib", "root")
__root_module = os.path.join(__root_path, "ROOT.py")
if not os.path.exists(__root_module):
    print "Could not find ROOT module"
    sys.exit(1)
    
if not __root_path in sys.path:
    sys.path.insert(0, __root_path)

import ROOT


class EventMonitor(object):
    """ This class shows an event
    """
    def __init__(self, name, module):
        """ Creates the canvas and the histograms
        """
        self.module = module
        self.cnvs = ROOT.TCanvas("cnvs_%s" % name, "Module %s" % module.id)
        self.hst1d = ROOT.TH1D("hst_%s" % module.id,
                               "Module %s Event Monitor" % module.id,
                               module.nchan,
                               - 0.5, module.nchan - 0.5
                               )
        if module.map:
            self.hst2d = ROOT.TH2D("hst_%s_map" % module.id,
                                   "Module %s - Event map" % module.id,
                                   module.nx, -0.5, module.nx - 0.5,
                                   module.ny, -0.5, module.ny - 0.5)
            self.cnvs.Divide(1, 2)
        else:
            self.hst2d = None
        self.cnvs.cd(1)
        self.hst1d.Draw()
        if self.hst2d:
            self.cnvs.cd(2)
            self.hst2d.Draw("colz")
        
    def add_event(self, evtnum, evt):
        """ This expects as input a list of (channel, value) tuples
        """
        self.hst1d.Reset()
        self.hst1d.SetTitle("Module %s - Event %d" % (self.module.id, evtnum))
        if self.hst2d:
            self.hst2d.Reset()
            self.hst2d.SetTitle("Module %s - Event %d" % (self.module.id, evtnum))
        
            
        for chan, value in evt:
            self.hst1d.Fill(chan, value)
            if self.hst2d:
                itm = self.module.map[chan]
                self.hst2d.Fill(itm.x, itm.y, value)
                
        self.cnvs.cd(1)
        self.hst1d.Draw()
        if self.hst2d:
            self.cnvs.cd(2)
            self.hst2d.Draw("colz")
            
        self.cnvs.Modified(True)
        self.cnvs.Update()


class EventManager(EventBuilder.EventBuilder):
    """ This class derives from EventBuilder
    """
    def __init__(self, pipe=None):
        """ Initialization
        """
        super(EventManager, self).__init__(pipe)
        self.monitor_list = {}

    def EndOfEvent(self, modid, evtnum):
        """ Called at the end of the event
        """
        super(EventManager, self).EndOfEvent(modid, evtnum)
        

    def append_data(self, module, evtnum, data):
        """ Called to append data from a given module
        """
        monitor = None
        if module.id in self.monitor_list:
            monitor = self.monitor_list[module.id]
        else:
            monitor = EventMonitor(module.id, module)
            self.monitor_list[module.id] = monitor

        
        monitor.add_event(evtnum, data)
        

            
class VATAmodule(vmedaq.VATAGP):
    """ Monitors a VATAGP module
    """
    def __init__(self, dtype, mngr):
        """ Initialization
        """
        super(VATAmodule, self).__init__(dtype, 0)
        self.manager = mngr
        self.hst_init = False
        
    def onNewData(self, module, evtcnt, data):
        """ Get new data
        """
#        if not self.monitor:
#            self.monitor = EventMonitor("VATAmodule", module)
#            self.monitor.hst1d.SetMaximum(1024.0/2.)
#            self.monitor.hst1d.SetMinimum(-100.0)
#            if self.monitor.hst2d:
#                self.monitor.hst2d.SetMaximum(1024.0/2.0)
#                self.monitor.hst2d.SetMinimum(-100.0)
                
            
        evt = []
        for chan, val, adj in data:
            try:
                signal = val - module.pedestals[chan]
                evt.append((chan, signal))
            except IndexError:
                pass
                
        self.manager.append_data(module, evtcnt, evt)
        if not self.hst_init:
            self.hst_init = True
            mon = self.manager.monitor_list[module.id]
            mon.hst1d.SetMaximum(1024.0/2.)
            mon.hst1d.SetMinimum(-100.0)
            if mon.hst2d:
                mon.hst2d.SetMaximum(1024.0/2.0)
                mon.hst2d.SetMinimum(-100.0)
        
        
def main(ifile=None):
    ROOT.gROOT.SetStyle("Plain")
    ROOT.gStyle.SetOptStat(0)
    ROOT.gStyle.SetPalette(1)
    file_list = None
    ped_file = None
    
    parser = OptionParser()
    parser.add_option("--pedestal",
                      dest=ped_file, action="store", type="string",
                      help="Pedestal file",
                      default=None)
    (options, args) = parser.parse_args()

    file_list = [ wfile for wfile in args ]
    if ifile:
        file_list.append(ifile)

    evt_mngr = EventManager()
    vata = VATAmodule(vmedaq.NormalData, evt_mngr)
    V = vmedaq.VMEdaq()
    V.register(vata)
    V.register(evt_mngr)
    V.open(file_list)
    if hasattr(options, 'ped_file'):
        print "pedestal file is", options.ped_file

    try:
        V.loop()
    except vmedaq.VMEdaqError, e:
        print "Something went wrong:", e
    except KeyboardInterrupt:
        print 'Arrrrgggg !!!!!' 


if __name__ == "__main__":
    main()
